# -*- coding: utf-8 -*-

from pathlib import Path
from typing import Any, Literal, cast

from openai import NOT_GIVEN
from openai.types import ImageModel
from sinapsis_core.data_containers.data_packet import DataContainer, ImagePacket
from sinapsis_core.template_base.dynamic_template import WrapperEntryConfig
from sinapsis_core.template_base.dynamic_template_factory import make_dynamic_template
from sinapsis_core.template_base.template import Template, TemplateAttributes
from sinapsis_core.utils.env_var_keys import SINAPSIS_BUILD_DOCS

from sinapsis_openai.helpers.openai_env_var_keys import OPENAI_API_KEY
from sinapsis_openai.helpers.openai_keys import OpenAIKeys
from sinapsis_openai.templates.openai_base import OpenAIBase, OpenAICreateType


class OpenAIImageCreation(OpenAIBase):
    """
    Template that wraps the `OpenAI.images.generate` API endpoint to
    perform image generation tasks. This template excludes the `prompt`,
    `model`, and `response_format` attributes from the API caller and are
    set through static template attributes. In this way, the `prompt` is received
    as input (via the content of a packet) and the `model` and `response_format` are defined in
    the `AttributesBaseModel`.

    Usage example:

    agent:
      name: my_test_agent
    templates:
    - template_name: InputTemplate
      class_name: InputTemplate
      attributes: {}
    - template_name: OpenAIImageCreationWrapper
      class_name: OpenAIImageCreationWrapper
      template_input: InputTemplate
      attributes:
        model: 'dall-e-2'
        response_format: url
        openai_init:
          api_key: openai_api_key
          max_retries: 2
        generate:
          n: 1

    """

    WrapperEntry = WrapperEntryConfig(
        wrapped_object=OpenAIBase.CLIENT,
        template_name_suffix="ImageCreationWrapper",
        additional_callables_to_inspect=[
            WrapperEntryConfig(
                wrapped_object=OpenAIBase.CLIENT(api_key=OPENAI_API_KEY).images.generate,
                exclude_method_attributes=[OpenAIKeys.prompt, OpenAIKeys.model, OpenAIKeys.response_format],
            )
        ],
    )

    class AttributesBaseModel(TemplateAttributes):
        """Attributes for the OpenAIImageCreation.

        Attributes:
            model (ImageModel): The image model to use for
            generating images.
            response_format (Literal["url", "b64_json"]):
                Specifies the format of the API response (default is "url").
        """

        model: ImageModel
        response_format: Literal["url", "b64_json"] = "url"

    def parse_results(self, responses: list | dict | Any, container: DataContainer) -> DataContainer:
        """
        For each of the responses, creates an ImagePacket to be
        appended to the DataContainer

        Args:
            responses (list[str]): List of URL from the OpenAI responses.
            container (DataContainer): The original data container.

        Returns:
            DataContainer: The modified container with appended ImagePackets.
        """
        for response in responses:
            container.images.append(
                ImagePacket(
                    content=response,
                    source=self.instance_name,
                )
            )
        return container

    def return_create_response(self, content: str | list | bytes) -> OpenAICreateType:
        """
        Sends an image generation request to the API endpoint
        using the provided prompt.

        Args:
            content (str | list | bytes): The prompt extracted from the packet.

        Returns:
            OpenAICreateType: The raw response from the OpenAI image generation API.
        """
        content = cast(str, content)
        return self.create(
            prompt=content,
            model=self.attributes.model,
            response_format=self.attributes.response_format,
            **self.attributes.generate.model_dump(),
        )


class OpenAIImageEdition(OpenAIImageCreation):
    """
    Template that wraps the `OpenAI.images.edit` API endpoint to perform
    image edition tasks.
    This template excludes the `image`, `mask`, `response_format`, `model`,
    and `prompt` attributes from the API caller.
    Instead, the image is provided through the attribute `path_to_image`
    and an optional mask through `path_to_mask` in the `AttributesBaseModel`.
    Additionally, the original image must be in `png` format,
    and, if a mask is provided, it must have the same dimensions as the original
    image.
    The PACKAET_TYPE_NAME indicates that this template processes packets stored in the
    text field of the DataContainer as the prompt

    Usage example:

    agent:
      name: my_test_agent
    templates:
    - template_name: InputTemplate
      class_name: InputTemplate
      attributes: {}
    - template_name: OpenAIImageEditionWrapper
      class_name: OpenAIImageEditionWrapper
      template_input: InputTemplate
      attributes:
        model: dall-e-2
        response_format: url
        path_to_mask' '/path/to/mask/image'
        path_to_image: '/path/to/image/to/be/edited'
        openai_init:
          api_key: openai_api_key
          max_retries: 2
        edit:
          n: 1
    """

    class AttributesBaseModel(OpenAIImageCreation.AttributesBaseModel):
        """
        Configuration attributes for the OpenAIImageEdition template.

        Attributes:
            model (Literal["dall-e-2"]): The fixed image model used for
            image edition tasks.
            path_to_mask (str | None): Optional file path to a mask image.
            If provided, the mask must have the same dimensions as the
            original image.
            path_to_image (str): The file path of the original image to
            be edited. The image is expected to be in `png` format.
        """

        model: Literal["dall-e-2"] = "dall-e-2"
        path_to_mask: str | None = None
        path_to_image: str

    WrapperEntry = WrapperEntryConfig(
        wrapped_object=OpenAIBase.CLIENT,
        template_name_suffix="ImageEditionWrapper",
        additional_callables_to_inspect=[
            WrapperEntryConfig(
                wrapped_object=OpenAIBase.CLIENT(api_key=OPENAI_API_KEY).images.edit,
                exclude_method_attributes=[
                    OpenAIKeys.image,
                    OpenAIKeys.mask,
                    OpenAIKeys.response_format,
                    OpenAIKeys.model,
                    OpenAIKeys.prompt,
                ],
            )
        ],
    )
    PACKET_TYPE_NAME = "texts"

    def return_create_response(self, content: str | list | bytes) -> OpenAICreateType:
        """
        Sends an image editing request to the API endpoint with the
        provided prompt.
        The image is taken from `path_to_image` and an optional
         mask from `path_to_mask` (if provided).
        These values are supplied exclusively via template attributes,
        which is why the corresponding API parameters are excluded.
        It is also expected that the image is in `png` format and,
        if a mask is used, that the mask matches the size of the
        original image.

        Args:
            content (str | list | bytes): The prompt for image editing,
            extracted from the packet.

        Returns:
            OpenAICreateType: The raw response from the OpenAI image editing API.
        """
        mask = Path(self.attributes.path_to_mask) if self.attributes.path_to_mask else NOT_GIVEN
        return self.create(
            image=Path(self.attributes.path_to_image),
            mask=mask,
            prompt=content,
            response_format=self.attributes.response_format,
            **self.attributes.edit.model_dump(),
        )


def __getattr__(name: str) -> Template:
    """
    Only create a template if it's imported, this avoids creating all
    the base models for all templates and potential import errors due
    to not available packages.
    """
    if name in OpenAIImageCreation.WrapperEntry.module_att_names:
        return make_dynamic_template(name, OpenAIImageCreation)
    if name in OpenAIImageEdition.WrapperEntry.module_att_names:
        return make_dynamic_template(name, OpenAIImageEdition)


__all__ = OpenAIImageCreation.WrapperEntry.module_att_names + OpenAIImageEdition.WrapperEntry.module_att_names

if SINAPSIS_BUILD_DOCS:
    dynamic_templates = [__getattr__(template_name) for template_name in __all__]
    for template in dynamic_templates:
        globals()[template.__name__] = template
        del template



# -*- coding: utf-8 -*-

from typing import Any, cast

from openai.types import ChatModel
from sinapsis_core.data_containers.data_packet import DataContainer, Packet, TextPacket
from sinapsis_core.template_base.dynamic_template import WrapperEntryConfig
from sinapsis_core.template_base.dynamic_template_factory import make_dynamic_template
from sinapsis_core.template_base.template import Template, TemplateAttributes
from sinapsis_core.utils.env_var_keys import SINAPSIS_BUILD_DOCS

from sinapsis_openai.helpers.openai_env_var_keys import OPENAI_API_KEY
from sinapsis_openai.helpers.openai_keys import OpenAIKeys
from sinapsis_openai.templates.openai_base import OpenAIBase, OpenAICreateType


class OpenAIChatCompletion(OpenAIBase):
    """
    OpenAIChatCompletion is a specialized template that wraps the
    'OpenAI.chat.completions.create' method.
    The template is built upon the common functionality provided by OpenAIBase
    to transform input data packets into appropriately formatted message lists,
    send the chat request to the OpenAI API, and parse the resulting responses
    into new text packets. This template requires configuration of both the API
    parameters (via the openai_init attribute) and the chat-specific settings
    (for example, the ChatModel to be used).

    The class attribute 'WrapperEntry' is defined using a nested WrapperEntryConfig.
    This setup enables dynamic template generation by first wrapping the base OpenAI
    client and then, in a nested manner, wrapping the specific OpenAI API call
    (chat.completions.create). In this way, only the desired method parameters
    (excluding those such as messages, audio, and model) are exposed for configuration.

    Usage example:
    agent:
      name: my_test_agent
    templates:
    - template_name: InputTemplate
      class_name: InputTemplate
      attributes: {}
    - template_name: OpenAIChatWrapper
      class_name: OpenAIChatWrapper
      template_input: InputTemplate
      attributes:
        model: gpt-4o
        openai_init:
          api_key: openai_api_key
          max_retries: 2
        create:
          max_tokens: 1000
          top_p: 3
    """

    WrapperEntry = WrapperEntryConfig(
        wrapped_object=OpenAIBase.CLIENT,
        template_name_suffix="ChatWrapper",
        additional_callables_to_inspect=[
            WrapperEntryConfig(
                wrapped_object=OpenAIBase.CLIENT(api_key=OPENAI_API_KEY).chat.completions.create,
                exclude_method_attributes=[OpenAIKeys.messages, OpenAIKeys.audio, OpenAIKeys.model],
            )
        ],
    )

    class AttributesBaseModel(TemplateAttributes):
        """AttributesBaseModel for OpenAIChatCompletion.
        Attributes:
            model (ChatModel): The chat model to be used for the completion.
        """

        model: ChatModel

    @staticmethod
    def unpack_packet_content(packet: Packet) -> list | str:
        """
        Transforms the content of the given Packet into a format required by the OpenAI chat API. The method
        wraps the packet's text content into a list containing a single dictionary that specifies the role ("user")
        and the content of the message.

        Args:
            packet (Packet): A data packet containing the input text message.

        Returns:
            list | str: A list containing a dictionary with keys "role" and "content",
            which represents the chat message.
        """
        message_list = [
            {
                "role": "user",
                "content": packet.content,
            }
        ]
        return message_list

    def return_create_response(self, content: list | str | bytes) -> OpenAICreateType:
        """
        Sends a request to the OpenAI chat completions endpoint by invoking the wrapped callable with the transformed
        message content, selected chat model, and additional parameters specified in the 'create' attribute. The method
        casts the content to a list (of messages) and issues the API call, returning the raw API response.

        Args:
            content (list | str | bytes): The processed input payload (a list of chat messages) to be sent to the API.

        Returns:
            OpenAICreateType: The raw response received from the OpenAI chat completions endpoint.
        """
        content = cast(list, content)
        return self.create(messages=content, model=self.attributes.model, **self.attributes.create.model_dump())

    def parse_results(self, responses: str | dict | Any, container: DataContainer) -> DataContainer:
        """
        Processes the list of responses from the OpenAI API by converting each response into a TextPacket, setting the
        packet's source to the instance name, and appending the packet to the provided DataContainer. This method
        ensures that the API responses are properly encapsulated for DataContainer compatibility.

        Args:
            responses (list[str]): A list of response texts from OpenAI.
            container (DataContainer): The container where the results will be stored.

        Returns:
            DataContainer: The updated container with parsed results.
        """

        for response in responses:
            text_packet = TextPacket(content=response, source=self.instance_name)
            container.texts.append(text_packet)
        return container


def __getattr__(name: str) -> Template:
    """
    Only create a template if it's imported, this avoids creating all the base models for all templates
    and potential import errors due to not available packages.
    """
    if name in OpenAIChatCompletion.WrapperEntry.module_att_names:
        return make_dynamic_template(name, OpenAIChatCompletion)


__all__ = OpenAIChatCompletion.WrapperEntry.module_att_names

if SINAPSIS_BUILD_DOCS:
    dynamic_templates = [__getattr__(template_name) for template_name in __all__]
    for template in dynamic_templates:
        globals()[template.__name__] = template
        del template



# -*- coding: utf-8 -*-
"""OpenAI.chat template module"""

from typing import Any, Union

from openai import APIConnectionError, BadRequestError, OpenAI
from openai.types import ImagesResponse
from openai.types.audio import Transcription, Translation
from openai.types.chat import ChatCompletion
from openai.types.embedding import Embedding
from openai.types.image import Image
from sinapsis_core.data_containers.data_packet import DataContainer, Packet
from sinapsis_core.template_base.dynamic_template import (
    BaseDynamicWrapperTemplate,
    WrapperEntryConfig,
)
from sinapsis_core.template_base.template import TemplateAttributeType

OpenAICreateType = Union[ChatCompletion, Transcription, Image, Embedding, Translation, str, Any, ImagesResponse]


class OpenAIBase(BaseDynamicWrapperTemplate):
    """
    OpenAIBase is a base class for creating a wrapper around the OpenAI API.
    It uses the `BaseDynamicWrapperTemplate` as its base class and
    initializes an instance of the `OpenAI`
    client when it is instantiated.


    Attributes:
        CLIENT (OpenAI): An instance of the OpenAI client.
        PACKET_TYPE_NAME (str): The name of the packet type for text data.
        WrapperEntry (WrapperEntryConfig): A configuration for the wrapped object.

    """

    CLIENT = OpenAI
    PACKET_TYPE_NAME = "texts"

    WrapperEntry = WrapperEntryConfig(wrapped_object=CLIENT)

    def __init__(self, attributes: TemplateAttributeType) -> None:
        super().__init__(attributes)
        self.openai = self.CLIENT(**self.attributes.openai_init.model_dump())
        self.create = self.wrapped_callable

    @staticmethod
    def unpack_packet_content(packet: Packet) -> list | str:
        """
        Unpacks the raw content from a given Packet instance.
        This method is intended to extract the core message content which will
        then be sent to the OpenAI API. The default implementation simply returns
        the Packet's content, although subclasses may override this behavior if a
        more complex transformation is required.

        Args:
            packet (Packet): Packet to unpack.

        Returns:
            Any: The unpacked content.
        """
        return packet.content

    def get_results(self, results: OpenAICreateType) -> str | dict | Any:
        """
        Processes and extracts relevant data from the OpenAI API response.
        Depending on the type of the response, appropriate fields are extracted.
        In the case of ChatCompletion, the response to the message content is
        returned; while for AudioTranscription and/or AudioTranslation, the
        updated Audio is returned.

        Args:
            results (OpenAICreateType): The raw response object returned by an
            OpenAI API call.

        Returns:
            str | dict | Any: Processed response data
            (text for chat/audio, list for images, etc.).
        """
        if isinstance(results, ChatCompletion):
            return results.choices[0].message.content if results.choices else ""

        elif isinstance(results, Transcription | Translation):
            return results.text

        elif isinstance(results, ImagesResponse):
            if results.data:
                if self.attributes.response_format == "url":
                    return results.data[0].url
                else:
                    return results.data[0].base64_json
            return ""
        elif isinstance(results, Embedding):
            return results.embedding

        else:
            return results

    def return_create_response(self, content: list | str | bytes) -> OpenAICreateType:
        """
        Constructs and returns the API response by invoking the wrapped
        creation callable with the provided content.
        This method must be implemented by specialized templates in order
        to correctly map input parameters to
        the OpenAI API request parameters as required by the particular endpoint.

        Args:
            content (list | str | bytes): Content to be passed as input to the API

        Returns:
            OpenAICreateType: The response from the OpenAI API.
        """

    def generate_response_from_client(self, packet: Packet) -> OpenAICreateType | None:
        """
        Extracts content from a given packet, sends the content to the
        OpenAI API using the wrapped callable, and then returns the
        processed result. This method catches API-specific connection
        errors or bad requests, logs a warning and returns None
        to signal the failure.

        Args:
            packet (Packet): Packet in the container to be passed to the client.

        Returns:
            OpenAICreateType | None: Processed results from the API call
            if successful. In case of an APIConnectionError or BadRequestError,
            the method logs the error with warning and returns None.

        Raises:
            APIConnectionError: If there is a connection issue contacting the API.
            BadRequestError: If the API returns an error due to an invalid request.
        """
        message = self.unpack_packet_content(packet)
        try:
            results = self.return_create_response(message)

        except (APIConnectionError, BadRequestError) as err:
            self.logger.warning(f"Error processing request: {err}")
            results = None

        return results

    def parse_results(self, responses: str | dict | Any, container: DataContainer) -> DataContainer:
        """
        Parses the responses obtained from the OpenAI API call and
        appropriately packages them into new data packets that are
        appended to the provided DataContainer.
        The method is responsible for converting raw API responses
        into a format suitable for DataContainers.
        Args:
            responses (str | dict | Any): Result from the call to
            the OpenAI API
            container (DataContainer): DataContainer where the result
            will be inserted as a Packet
        Returns:
            DataContainer: The updated data container with the newly appended packets
        """

    def execute(self, container: DataContainer) -> DataContainer:
        """
        Executes the overall process of retrieving and processing
        OpenAI API responses. It iterates over each target packet in the DataContainer,
        sends the content of, so it can be processed by the API. Then, the response is
        gathered and processed using get_results. Finally, the processed responses are
        added to the container for downstream usage.

        Args:
            container (DataContainer): The container holding the input data packets.

        Returns:
            DataContainer: The updated container with responses from OpenAI.
        """
        data_packet = getattr(container, self.PACKET_TYPE_NAME)
        responses_from_openai: list[str | dict | Any] = []
        for packet in data_packet:
            response_output = self.generate_response_from_client(packet)
            responses_from_openai.append(self.get_results(response_output))

        container = self.parse_results(responses_from_openai, container)

        return container
